package com.company.classicaudio.listener;

import com.company.classicaudio.entity.Instrument;
import com.company.classicaudio.entity.User;
import io.jmix.core.*;
import io.jmix.core.event.AttributeChanges;
import io.jmix.core.event.EntityChangedEvent;
import io.jmix.core.security.CurrentAuthentication;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.event.EventListener;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Component;

import java.util.Set;

@Component
public class InstrumentEventListener {
    private final CurrentAuthentication currentAuthentication;
    @Autowired
    private DataManager dataManager;
    @Autowired
    private Messages messages;
    @Autowired
    private FileStorage fileStorage;

    public InstrumentEventListener(CurrentAuthentication currentAuthentication) {
        this.currentAuthentication = currentAuthentication;
    }

    @EventListener
    public void onInstrumentChangedBeforeCommit(final EntityChangedEvent<Instrument> event) {
        Instrument instrument;
        FileRef oldAudioFileRef = null;
        long oldAudioFileSize = 0;
        long previousSize = 0;
        long sizediff = 0;
        long storageRemaining;
        final String instrumentFetchPlan = "_base";
        final String membershipFetchPlan = "_base";

        if (event.getType() != EntityChangedEvent.Type.CREATED) {
            AttributeChanges attributeChanges = event.getChanges();
            Set<String> attributes = attributeChanges.getAttributes();
            for (String attribute : attributes) {
                if (attribute.equals("audioFile")) {
                    oldAudioFileRef = attributeChanges.getOldValue("audioFile");
                }
                if (attribute.equals("audioFileSize")) {
                    oldAudioFileSize = attributeChanges.getOldValue("audioFileSize");
                }
            }
        }

        if (event.getType() == EntityChangedEvent.Type.CREATED) {
            instrument = dataManager.load(event.getEntityId())
                    .fetchPlan(instrumentFetchPlan)
                    .one();
            if (instrument.getAudioFile() != null) {   // was an audio file added during creation
                sizediff = instrument.getAudioFileSize();
            }
        }

        if (oldAudioFileRef != null) {   // audio file was removed or exchanged
            previousSize = oldAudioFileSize;
            deleteAudioFile(oldAudioFileRef);    // permanently remove the old audio file
        }

        if (event.getType() == EntityChangedEvent.Type.UPDATED) {
            // check for new audio file and get its size
            instrument = dataManager.load(event.getEntityId())
                    .fetchPlan(instrumentFetchPlan)
                    .one();
            if (oldAudioFileRef != null) {
                if (instrument.getAudioFile() != null) {
                    sizediff = instrument.getAudioFileSize() - previousSize;
                } else {
                    sizediff = previousSize * -1;
                }
            } else if (instrument.getAudioFile() != null) {
                sizediff = instrument.getAudioFileSize();    // first-time audio file addition
            }
        }

 /*     In my real application each user has a Membership object that describes their available resources dependent upon their subscription.
        I use this code and the code above and farther below, that is also commented out, to store the amount of storage that the user still has.
        You may not need this, but it as an example.
        Membership membership = dataManager.load(Membership.class)
                .query("select e from nf_Membership e where e.creator.username = :username")
                .parameter("username", getCurrentUser().getUsername())
                .fetchPlan(membershipFetchPlan)
                .optional().orElse(null);

        if (membership == null) {   // this should only happen if the Admin user is deleted before the creator
            return;
        }
*/
        if (event.getType() == EntityChangedEvent.Type.DELETED) {
            Id<User> creatorId = event.getChanges().getOldReferenceId("creator");
            if (creatorId != null) {
                if (!creatorId.getValue().equals(getCurrentUser().getId())) {
                    return; // the user is being deleted, the membership information is no longer relevant
                }
            }
            if (oldAudioFileRef != null) {
                sizediff = previousSize * -1;
            }
        }

/*      See the comment above.
        membership.setStorageUsed(membership.getStorageUsed() + sizediff);
        storageRemaining = membership.getStorageLimit() - membership.getStorageUsed();
        if (storageRemaining < 0) {
            storageRemaining = 0;
        }
        membership.setStorageRemaining(storageRemaining);

        dataManager.save(membership);
*/
    }

    private void deleteAudioFile(FileRef fileRef) {
        final Logger log = LoggerFactory.getLogger(InstrumentEventListener.class);

        try {
            fileStorage.removeFile(fileRef);
        } catch (FileStorageException e) {
            String errorMsg = messages.getMessage("cannotRemoveAudioFileFromFileStorage");
            log.error(errorMsg, e);
            throw new RuntimeException(errorMsg, e);
        }
    }

    public User getCurrentUser() {
        User currentUser;

        UserDetails userDetails = currentAuthentication.getUser();
        if (userDetails instanceof User) {
            currentUser = (User) userDetails;
        } else {
            currentUser = dataManager.load(User.class)
                    .query("select u from USER_ u where u.username = :username")
                    .parameter("username", userDetails.getUsername())
                    .one();
        }
        return(currentUser);
    }
}