package com.company.classicaudio.screen.instrument;

import io.jmix.core.*;
import io.jmix.ui.Dialogs;
import io.jmix.ui.Notifications;
import io.jmix.ui.action.Action;
import io.jmix.ui.action.DialogAction;
import io.jmix.ui.model.InstanceContainer;
import io.jmix.ui.screen.*;
import io.jmix.ui.component.*;
import io.jmix.ui.upload.TemporaryStorage;

import com.company.classicaudio.entity.Instrument;

import com.vaadin.server.StreamResource;
import com.vaadin.ui.AbstractOrderedLayout;
import com.vaadin.ui.Audio;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.util.Objects;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.springframework.beans.factory.annotation.Autowired;

@UiController("Instrument.edit")
@UiDescriptor("instrument-edit.xml")
@EditedEntityContainer("instrumentDc")
public class InstrumentEdit extends StandardEditor<Instrument> {

    protected Audio instrumentAudio;

    private FileRef newAudioFileRef = null;
    private StreamResource instrumentAudioResource;

    private long storageRemaining;
    private long currentFileSize = 0;   // existing file size when first opening this screen
    private long newFileSize = 0;       // new file size after upload in this screen

    private final Logger log = LoggerFactory.getLogger(InstrumentEdit.class);

    @Autowired
    private Dialogs dialogs;
    @Autowired
    private Notifications notifications;
    @Autowired
    private Messages messages;

    @Autowired
    private InstanceContainer<Instrument> instrumentDc;
    @Autowired
    private TextField<Long> fileSizeTextField;
    @Autowired
    private TextField<Long> storageRemainingTextField;
    @Autowired
    private FileStorageUploadField instrumentAudioFileStorageUploadField;

    @Autowired
    private Button instrumentPauseAudioBtn;
    @Autowired
    private Button instrumentPlayAudioBtn;
    @Autowired
    private Button instrumentRewindAudioBtn;

    @Autowired
    private FileStorage fileStorage;
    @Autowired
    private TemporaryStorage temporaryStorage;


    @Subscribe
    public void onInit(InitEvent event) {

        instrumentAudio = new Audio();
        instrumentAudio.setShowControls(false);
        getWindow().unwrap(AbstractOrderedLayout.class).addComponent(instrumentAudio);
    }

    @Subscribe
    public void onAfterInit(AfterInitEvent event) {

        instrumentPlayAudioBtn.setEnabled(false);
        instrumentPauseAudioBtn.setEnabled(false);
        instrumentRewindAudioBtn.setEnabled(false);
    }

    @Subscribe
    public void onAfterShow(AfterShowEvent event) {

        storageRemaining = instrumentDc.getItem().getStorageRemaining();

        if (instrumentDc.getItem().getAudioFile() != null) {
            currentFileSize = instrumentDc.getItem().getAudioFileSize();
            instrumentAudioFileStorageUploadField.setValue(instrumentDc.getItem().getAudioFile());
            instrumentAudioResource = new StreamResource(
                    (StreamResource.StreamSource) () -> {
                        byte[] bytes = new byte[0];
                        try {
                            bytes = fileStorage.openStream(instrumentDc.getItem().getAudioFile()).readAllBytes();
                        } catch (IOException e) {
                            log.error("ERROR: Unable to load audio file - " + instrumentDc.getItem().getAudioFile().getFileName(), e);
                        }
                        return new ByteArrayInputStream(bytes);
                    }, "") {
                @Override
                public String getMIMEType() {
                    return "audio/mp3";
                }
            };
            instrumentAudio.setSource(instrumentAudioResource);
            instrumentPlayAudioBtn.setEnabled(true);
            instrumentAudioFileStorageUploadField.setShowClearButton(true);
        } else {
            instrumentAudioFileStorageUploadField.setShowClearButton(false);
        }

        fileSizeTextField.setValue(currentFileSize);

        storageRemainingTextField.setValue(storageRemaining);
        instrumentAudioFileStorageUploadField.setFileSizeLimit(storageRemaining + currentFileSize);
    }


    @Subscribe("instrumentAudioFileStorageUploadField")
    public void onInstrumentAudioFileStorageUploadFieldFileUploadStart(UploadField.FileUploadStartEvent event) {

        instrumentAudio.pause();
        instrumentAudio.setSource(null);

        instrumentPlayAudioBtn.setEnabled(false);
        instrumentPauseAudioBtn.setEnabled(false);
        instrumentRewindAudioBtn.setEnabled(false);
    }

    @Subscribe("instrumentAudioFileStorageUploadField")
    public void onInstrumentAudioFileStorageUploadFieldFileUploadSucceed(SingleFileUploadField.FileUploadSucceedEvent event) {

        File file = temporaryStorage.getFile(Objects.requireNonNull(instrumentAudioFileStorageUploadField.getFileId()));
        if (file != null) {
            notifications.create(Notifications.NotificationType.TRAY)
                    .withCaption(messages.getMessage("audioFileStoredSuccessfully") + " - " + file.getAbsolutePath())   // TODO remove path after testing
                    .show();
        } else {
            String failedMessage = messages.getMessage("audioFileStorageFailed") + " - " + event.getFileName();
            notifications.create(Notifications.NotificationType.ERROR)
                    .withCaption(failedMessage)
                    .show();
            log.error(failedMessage);
            return;
        }

        if (newAudioFileRef != null) {   // remove the last uploaded audio file during this session, since a new one was chosen
            fileStorage.removeFile(newAudioFileRef);
            newAudioFileRef = null;
        }

        instrumentDc.getItem().setAudioFileSize(file.length());
        FileRef newAudioFileRef = temporaryStorage.putFileIntoStorage(instrumentAudioFileStorageUploadField.getFileId(), event.getFileName());
        instrumentDc.getItem().setAudioFile(newAudioFileRef);
        instrumentAudioFileStorageUploadField.setValue(newAudioFileRef);

        fileSizeTextField.setValue(instrumentDc.getItem().getAudioFileSize());
        if (newFileSize == 0) { // then this is the first upload after opening this screen
            storageRemaining += currentFileSize;    // if currentFileSize > 0, then, since it will be replaced by the new upload, we give its size back
            currentFileSize = 0;
        } else {    // if a new file was uploaded after opening this screen, and it's now being replaced (maybe the wrong file was chosen), then we give its size back
            storageRemaining += newFileSize;
        }
        storageRemaining -= instrumentDc.getItem().getAudioFileSize();
        storageRemainingTextField.setValue(storageRemaining);
        instrumentDc.getItem().setStorageRemaining(storageRemaining);
        newFileSize = instrumentDc.getItem().getAudioFileSize();

        instrumentAudioResource = new StreamResource(
                (StreamResource.StreamSource) () -> {
                    byte[] bytes = new byte[0];
                    try {
                        bytes = fileStorage.openStream(instrumentDc.getItem().getAudioFile()).readAllBytes();
                    } catch (IOException e) {
                        log.error("ERROR: Unable to load audio file - " + instrumentDc.getItem().getAudioFile().getFileName(), e);
                    }
                    return new ByteArrayInputStream(bytes);
                }, "") {
            @Override
            public String getMIMEType() {
                return "audio/mp3";
            }
        };

        instrumentPlayAudioBtn.setEnabled(true);
        instrumentAudioFileStorageUploadField.setShowClearButton(true);
    }

    @Subscribe("instrumentAudioFileStorageUploadField")
    public void onInstrumentAudioFileStorageUploadFieldBeforeValueClear(SingleFileUploadField.BeforeValueClearEvent event) {
        event.preventClearAction();
    }

    @Subscribe("instrumentAudioFileStorageUploadField")
    public void onInstrumentAudioFileStorageUploadFieldAfterValueClear(SingleFileUploadField.AfterValueClearEvent event) {

        dialogs.createOptionDialog()
                .withCaption(messages.getMessage("com.company.classicaudio.screen.instrument", "clearAudioFileConfirmationCaption"))
                .withMessage(messages.getMessage("com.company.classicaudio.screen.instrument", "clearAudioFileConfirmationMessage"))
                .withActions(
                        new DialogAction(DialogAction.Type.YES)
                                .withHandler(e -> instrumentAudioFileStorageUploadField.clear()),
                        new DialogAction(DialogAction.Type.NO, Action.Status.PRIMARY)
                )
                .show();
    }

    @Subscribe("instrumentAudioFileStorageUploadField")
    public void onInstrumentAudioFileStorageUploadFieldValueChange(HasValue.ValueChangeEvent<FileRef> event) {

        if (event.getValue() == null) { // clear instrumentAudioFileStorageUploadField
            instrumentAudioFileStorageUploadField.setShowClearButton(false);
            instrumentAudio.pause();
            instrumentAudio.setSource(null);
            try {
                if (newAudioFileRef != null) {   // remove the last uploaded audio file for this session
                    fileStorage.removeFile(newAudioFileRef);
                    newAudioFileRef = null;
                }
            } catch (FileStorageException e) {
                // TODO move this text to the messages.properties
                log.error("Unable to remove the previous audio file - " + newAudioFileRef.getFileName());
            }
            instrumentDc.getItem().setAudioFile(null);
            if (currentFileSize != 0) {
                storageRemaining += currentFileSize;
                currentFileSize = 0;
            } else {
                storageRemaining += newFileSize;
            }
            newFileSize = 0;
            instrumentDc.getItem().setAudioFileSize(newFileSize);
            fileSizeTextField.setValue(newFileSize);
            storageRemainingTextField.setValue(storageRemaining); // reset to stored value
            instrumentDc.getItem().setStorageRemaining(storageRemaining);
            instrumentPlayAudioBtn.setEnabled(false);
            instrumentPauseAudioBtn.setEnabled(false);
            instrumentRewindAudioBtn.setEnabled(false);
        }
    }

    @Subscribe("instrumentPlayAudioBtn")
    public void onInstrumentPlayAudioBtnClick(Button.ClickEvent event) {

        instrumentAudio.setSource(instrumentAudioResource);
        instrumentAudio.play();
        instrumentPlayAudioBtn.setEnabled(false);
        instrumentPauseAudioBtn.setEnabled(true);
        instrumentRewindAudioBtn.setEnabled(false);
    }

    @Subscribe("instrumentPauseAudioBtn")
    public void onInstrumentPauseAudioBtnClick(Button.ClickEvent event) {

        instrumentAudio.pause();
        instrumentPlayAudioBtn.setEnabled(true);
        instrumentPauseAudioBtn.setEnabled(false);
        instrumentRewindAudioBtn.setEnabled(true);
    }

    @Subscribe("instrumentRewindAudioBtn")
    public void onInstrumentRewindAudioBtnClick(Button.ClickEvent event) {

        instrumentAudio.setSource(null);
        instrumentPauseAudioBtn.setEnabled(false);
        instrumentRewindAudioBtn.setEnabled(false);
    }
}